/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

    Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "VerInfo.h"

#include "shlobj.h"
#include "shlwapi.h"

#pragma comment(lib, "../xPaintClass_SDK/SDK/xPaintClass-1.0.lib")

static int iScreenX = GetSystemMetrics(SM_XVIRTUALSCREEN);
static int iScreenY = GetSystemMetrics(SM_YVIRTUALSCREEN);
static int iScreenWidth = GetSystemMetrics(SM_CXVIRTUALSCREEN);
static int iScreenHeight = GetSystemMetrics(SM_CYVIRTUALSCREEN);

static int iRealMonitors = GetSystemMetrics(SM_CMONITORS);

static BOOL bEnableDoubleClicks = GetxRCBool("xDesk", "EnableDoubleClicks");
static int iDoubleClickTimeOut = GetDoubleClickTime();

#define SHELL_MAGICNUMBER 0x49474541

#define LM_UPDATEBG (WM_USER + 1)

#define GET_X_LPARAM(lp) ((int)(short)LOWORD(lp))
#define GET_Y_LPARAM(lp) ((int)(short)HIWORD(lp))

//MouseWheel
//-----------------------------------------------------
#ifndef WM_MOUSEWHEEL
	#define WM_MOUSEWHEEL 0x020A
#endif
//-----------------------------------------------------

//XBUTTONS
//-----------------------------------------------------
#ifndef WM_XBUTTONDOWN
  #define WM_XBUTTONDOWN 0x020B
#endif
#ifndef WM_XBUTTONUP
  #define WM_XBUTTONUP 0x020C
#endif

#ifndef WM_XBUTTONDBLCLK
  #define WM_XBUTTONDBLCLK 0x020D
#endif

#ifndef MK_XBUTTON1  
   #define MK_XBUTTON1 0x0020
#endif
#ifndef MK_XBUTTON2  
   #define MK_XBUTTON2 0x0040
#endif
#ifndef XBUTTON1  
   #define XBUTTON1 0x0001
#endif
#ifndef XBUTTON2
  #define XBUTTON2 0x0002
#endif
//-----------------------------------------------------

class Region
{
public:

	Region() : 
		strOrgX("0"),
		strOrgY("0"),
		strOrgWidth("0"),
		strOrgHeight("0"),

		strAction("")
	{
		SetRectEmpty(&region);
	};

	virtual ~Region() {};

	string strOrgX;
	string strOrgY;
	string strOrgWidth;
	string strOrgHeight;
	
	RECT region;

	string strAction;
};

typedef vector<Region *>RegionList;
typedef vector<Region *>::iterator RegionListIterator;

struct ModClick
{
	int iModifier;
	string strAction;
};

typedef vector<ModClick *>ModClickList;
typedef vector<ModClick *>::iterator ModClickListIterator;

#define StringCchCopy(a, b, c) strncpy((a), (c), (b))
#define StringCchLength(a, b, c) *(c) = strlen((a)), S_OK
#define StringCchPrintf _snprintf

HINSTANCE hInstance = NULL;
HWND messageHandler = NULL;

HDC backupDC = NULL;
HBITMAP backupBitmap = NULL;
HBITMAP oldbackupBMP = NULL;

void PrepareImageEffects();

//Texture Painting Helpers
void RGB2HSL(unsigned char R,unsigned char G,unsigned char B,unsigned char &H,unsigned char &S,unsigned char &L);
void HSL2RGB(unsigned char &R,unsigned char &G,unsigned char &B,unsigned char H,unsigned char S,unsigned char L);

float ToRGB(float rm1, float rm2, float rh);

void RGB2HSLFloat(float R, float G, float B, float &H, float &S, float &L);
void HSL2RGBFloat(float &R, float &G, float &B, float H, float S, float L);

int SaveBitmap(HDC hdc, int width, int height, string strFilename);
LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue);

//-----------------------------------------------------------------
//Image
//-----------------------------------------------------------------

COLORREF m_clrHueColor = Getx2RCColor("xDesk", "HueColor", RGB(255,255,255));		
int m_iHueIntensity = Getx2RCInt("xDesk", "HueIntensity", 0, 0, 100);
int	m_iLuminanceIntensity = Getx2RCInt("xDesk", "LuminanceIntensity", 0, -100, 100);
int	m_iSaturationIntensity = Getx2RCInt("xDesk", "SaturationIntensity", 100, 0, 200);
COLORREF m_clrMixColor = Getx2RCColor("xDesk", "MixColor", RGB(255,255,255));						
int	m_iMixIntensity = Getx2RCInt("xDesk", "MixIntensity", 0, 0, 100);

void BangSetWorkArea(HWND hwndCaller, LPCSTR pszArgs);
void BangSetWallpaper(HWND hwndCaller, LPCSTR pszArgs);
void BangSetup(HWND hwndCaller, LPCSTR pszArgs);

HWND hxDeskWnd; // main window handle (desktop)
HWND hLitestepWnd; // Shell window handle

BOOL bIsFirst = TRUE;

BOOL bMouse1Pressed = FALSE;
BOOL bMouse2Pressed = FALSE;
BOOL bMouse3Pressed = FALSE;

BOOL bMouse1Disabled = FALSE;
BOOL bMouse2Disabled = FALSE;
BOOL bMouse3Disabled = FALSE;
BOOL bMouse4Disabled = FALSE;
BOOL bMouse5Disabled = FALSE;

BOOL bOverlaysUsed = FALSE;
BOOL bOverlaysInitialization = FALSE;

RECT rWorkArea1, rWorkArea2, rWorkArea3;

StringList GetxRCLineList(const string &uniqueKey);
void SetWorkArea();

int lsMessages[] = {
	LM_GETREVID,
	0
};

void onWheelDown();
void onWheelUp();

void onLButtonDblClk(int x, int y);
void onLButtonDown(int x, int y);
void onLButtonUp(int x, int y);
void onMButtonDblClk(int x, int y);
void onMButtonDown(int x, int y);
void onMButtonUp(int x, int y);
void onRButtonDblClk(int x, int y);
void onRButtonDown(int x, int y);
void onRButtonUp(int x, int y);
void onX1ButtonDown(int x, int y);
void onX1ButtonUp(int x, int y);
void onX1ButtonDblClk(int x, int y);
void onX2ButtonDown(int x, int y);
void onX2ButtonUp(int x, int y);
void onX2ButtonDblClk(int x, int y);

void ontemplateButtonDown(int x, int y, int actualMouseClick, RegionList &ClickRegions);
void ontemplateButtonUp(int x, int y, int actualMouseClick, RegionList &ClickRegions);

void DropExecute(string strAction, LONG wParam);

//Actions
ModClickList onDropList;

ModClickList wheelDownList;
ModClickList wheelUpList;

ModClickList leftModClickDownList;
ModClickList leftModClickUpList;
ModClickList leftModDoubleClickList;

ModClickList middleModClickDownList;
ModClickList middleModClickUpList;
ModClickList middleModDoubleClickList;

ModClickList rightModClickDownList;
ModClickList rightModClickUpList;
ModClickList rightModDoubleClickList;

ModClickList X1ModClickDownList;
ModClickList X1ModClickUpList;
ModClickList X1ModDoubleClickList;

ModClickList X2ModClickDownList;
ModClickList X2ModClickUpList;
ModClickList X2ModDoubleClickList;

void Fill_ModClickList(StringList inputList, ModClickList &outputList);
bool Check_ModClickList(ModClickList inputList, int iMouseButton = -1, LONG wParam = NULL);

RegionList LeftClickRegions;
RegionList LeftDoubleClickRegions;
	
RegionList MiddleClickRegions;
RegionList MiddleDoubleClickRegions;

RegionList RightClickRegions;
RegionList RightDoubleClickRegions;

RegionList X1ClickRegions;
RegionList X1DoubleClickRegions;

RegionList X2ClickRegions;
RegionList X2DoubleClickRegions;

void Fill_ClickRegionList(StringList inputList, RegionList &outputList);
bool Check_ClickRegionList(int x, int y, RegionList inputlist);

void OverlayPainter(HDC hDC);

void CleanupPointers();

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			if ( backupBitmap )
			{
				backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
				DeleteObject(backupBitmap);
			}
			if ( backupDC )
				DeleteDC(backupDC);
			if ( oldbackupBMP )
				DeleteObject(oldbackupBMP);

			backupDC = NULL;
			backupBitmap = NULL;
			oldbackupBMP = NULL;

			RedrawWindow(hxDeskWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW );

			return 0;
		}

		case WM_DISPLAYCHANGE:
		case WM_SETTINGCHANGE:
		{
			if ( bOverlaysUsed )
				PostMessage(hWnd, LM_UPDATEBG, 0, 0);

			return 0;
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

LRESULT WINAPI DesktopBackgroundProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case WM_WINDOWPOSCHANGING:
			{
				// Trap windowposchanging messages
				WINDOWPOS *c = (WINDOWPOS*)lParam;
				c->hwnd = hxDeskWnd;
				c->hwndInsertAfter = HWND_BOTTOM;
				c->x = iScreenX;
				c->y = iScreenY;
				c->cx = iScreenWidth;
				c->cy = iScreenHeight;
				c->flags = c->flags & ~SWP_NOZORDER; //not really needed, as I never send a command with it.
				c->flags |= SWP_NOACTIVATE|SWP_NOSENDCHANGING;
				return 0;
			}

			case WM_ERASEBKGND:
				{
					if ( !bOverlaysUsed || !bOverlaysInitialization )
					{
						PaintDesktop( (HDC)wParam );
						return 1;
					}
					else
					{
						return 1;
					}
				}

			case WM_PAINT:
				{
					if ( !bOverlaysUsed || !bOverlaysInitialization )
						break;

					PAINTSTRUCT ps;

					HDC hDC = BeginPaint( hWnd, &ps );

					OverlayPainter(hDC);

					EndPaint( hWnd, &ps );

					return true;
				}

		case WM_CLOSE:
			{
				// If someone trys to exit the desktop window, lets make it a windows shutdown.
				PostMessage(hLitestepWnd, LM_RECYCLE, 3, 0);
				return 0;
			}

		case WM_SYSCOMMAND:
			{
				switch (wParam)
				{
					//For using the standard alt+F4 to shutdown windows
					case SC_CLOSE:
						{
							PostMessage(hLitestepWnd, LM_RECYCLE, 3, 0);
							return 0;
						}

					default:
						break;
				}
			}
			break;

		case WM_TIMER:
			{
				if (bEnableDoubleClicks)
				{
					KillTimer(hxDeskWnd, wParam);
					switch (wParam)
					{
						case 0:
							KillTimer(hxDeskWnd, 1);
							KillTimer(hxDeskWnd, 2);
							KillTimer(hxDeskWnd, 3);
							KillTimer(hxDeskWnd, 4);
							KillTimer(hxDeskWnd, 5);

							bMouse1Disabled = FALSE;
							bMouse2Disabled = FALSE;
							bMouse3Disabled = FALSE;
							bMouse4Disabled = FALSE;
							bMouse5Disabled = FALSE;
							break;
						case 1:
							if (!bMouse1Disabled)
								Check_ModClickList(leftModClickUpList, -10);
							bMouse1Disabled = FALSE;
							break;
						case 2:
							if (!bMouse2Disabled)
								Check_ModClickList(rightModClickUpList, -20);
							bMouse2Disabled = FALSE;
							break;
						case 3:
							if (!bMouse3Disabled)
								Check_ModClickList(middleModClickUpList, -30);
							bMouse3Disabled = FALSE;
							break;
						case 4:
							if (!bMouse4Disabled)
								Check_ModClickList(X1ModClickUpList);
							bMouse4Disabled = FALSE;
							break;
						case 5:
							if (!bMouse5Disabled)
								Check_ModClickList(X2ModClickUpList);
							bMouse5Disabled = FALSE;
							break;
					}
					return 0;
				}
			}
			break;
		
		case WM_DROPFILES:
			{
				if (onDropList.empty())
					break;

				Check_ModClickList(onDropList, 0, wParam);
				
				return true;
			}

		case WM_MOUSEWHEEL:
			{
				if ((short)(HIWORD(wParam)) < 0)
				{
					onWheelDown();

					return true;
				}
				else
				{
					onWheelUp();

					return true;
				}
			}

		case WM_LBUTTONDOWN:
		{	
			if ( !bMouse2Pressed && !bMouse3Pressed )
				bMouse1Pressed = TRUE;

			if ( bEnableDoubleClicks )
				KillTimer(hxDeskWnd, 1);

			onLButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			if ( bEnableDoubleClicks )
				SetTimer(hxDeskWnd, 1, iDoubleClickTimeOut, (TIMERPROC)NULL);
			else
				onLButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));

			bMouse1Pressed = FALSE;
			return true;
		}
		
		case WM_LBUTTONDBLCLK:
		{
			if ( bEnableDoubleClicks )
			{
				KillTimer(hxDeskWnd, 1);
				bMouse1Disabled = TRUE;
				SetTimer(hxDeskWnd, 0, 100, (TIMERPROC)NULL);
			}

			onLButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			if ( !bMouse1Pressed && !bMouse2Pressed )
				bMouse3Pressed = TRUE;

			if ( bEnableDoubleClicks )
				KillTimer(hxDeskWnd, 3);

			onMButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			if ( bEnableDoubleClicks )
				SetTimer(hxDeskWnd, 3, iDoubleClickTimeOut, (TIMERPROC)NULL);
			else
				onMButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));

			bMouse3Pressed = FALSE;
			return true;
		}

		case WM_MBUTTONDBLCLK:
		{
			if ( bEnableDoubleClicks )
			{
				KillTimer(hxDeskWnd, 3);
				bMouse3Disabled = TRUE;
				SetTimer(hxDeskWnd, 0, 100, (TIMERPROC)NULL);
			}

			onMButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			if ( !bMouse1Pressed && !bMouse3Pressed )
				bMouse2Pressed = TRUE;

			if ( bEnableDoubleClicks )
				KillTimer(hxDeskWnd, 2);

			onRButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_RBUTTONUP:
		{
			if ( bEnableDoubleClicks )
				SetTimer(hxDeskWnd, 2, iDoubleClickTimeOut, (TIMERPROC)NULL);
			else
				onRButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));

			bMouse2Pressed = FALSE;
			return true;
		}
		
		case WM_RBUTTONDBLCLK:
		{
			if ( bEnableDoubleClicks )
			{
				KillTimer(hxDeskWnd, 2);
				bMouse2Disabled = TRUE;
				SetTimer(hxDeskWnd, 0, 100, (TIMERPROC)NULL);
			}

			onRButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_XBUTTONDOWN:
		{
			if (HIWORD(wParam) == XBUTTON1)
			{
				if ( bEnableDoubleClicks )
					KillTimer(hxDeskWnd, 4);
		
				onX1ButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
			else if (HIWORD(wParam) == XBUTTON2)
			{
				if ( bEnableDoubleClicks )
					KillTimer(hxDeskWnd, 5);

				onX2ButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
						
			return true;
		}

		case WM_XBUTTONUP:
		{
			if (HIWORD(wParam) == XBUTTON1)
			{
				if ( bEnableDoubleClicks )
					SetTimer(hxDeskWnd, 4, iDoubleClickTimeOut, (TIMERPROC)NULL);
				else
					onX1ButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
			else if (HIWORD(wParam) == XBUTTON2)
			{
				if ( bEnableDoubleClicks )
					SetTimer(hxDeskWnd, 5, iDoubleClickTimeOut, (TIMERPROC)NULL);
				else
					onX2ButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
						
			return true;
		}

		case WM_XBUTTONDBLCLK:
		{
			if (HIWORD(wParam) == XBUTTON1)
			{
				if ( bEnableDoubleClicks )
				{
					KillTimer(hxDeskWnd, 4);
					bMouse4Disabled = TRUE;
					SetTimer(hxDeskWnd, 0, 100, (TIMERPROC)NULL);
				}

				onX1ButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
			else if (HIWORD(wParam) == XBUTTON2)
			{
				if ( bEnableDoubleClicks )
				{
					KillTimer(hxDeskWnd, 5);
					bMouse5Disabled = TRUE;
					SetTimer(hxDeskWnd, 0, 100, (TIMERPROC)NULL);
				}

				onX2ButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}

			return true;
		}

		case WM_CHILDACTIVATE:
		case WM_NCACTIVATE:
		case WM_SETFOCUS:
		case WM_KILLFOCUS:
		case WM_ACTIVATEAPP:
		case WM_ACTIVATE:
		case WM_PARENTNOTIFY:
			{
				SetWindowPos(hxDeskWnd,HWND_BOTTOM,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
				break;
			}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);

	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xDeskMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	WNDCLASS wcDesk;

	memset(&wcDesk,0,sizeof(wcDesk));
	wcDesk.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcDesk.lpfnWndProc = DesktopBackgroundProc;
	wcDesk.hInstance = hInstance;
	wcDesk.lpszClassName = "DesktopBackgroundClass";
	wcDesk.style = (bEnableDoubleClicks ? CS_DBLCLKS : 0);

	RegisterClass(&wcDesk);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xDeskMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	hxDeskWnd = CreateWindowEx(WS_EX_TOOLWINDOW,
		"DesktopBackgroundClass",
		"",
		WS_POPUP|WS_CLIPSIBLINGS|WS_CLIPCHILDREN,
		0,0,0,0,
		0,
		0,
		hInstance,
		0);
	
	if (!messageHandler)
		return 1;

	if (!hxDeskWnd)
		return 1;

	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;
	::hLitestepWnd = hParent;

	AddBangCommand("!xDeskSetWorkArea", BangSetWorkArea);
	AddBangCommand("!xDeskSetWallpaper", BangSetWallpaper);
	AddBangCommand("!xDesk", BangSetup);

	//Wheel
	Fill_ModClickList(GetxRCLineList("*xDeskOnWheelDown"), wheelDownList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnWheelUp"), wheelUpList);

	//Left
	Fill_ModClickList(GetxRCLineList("*xDeskOnLeftClickUp"), leftModClickUpList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnLeftClickDown"), leftModClickDownList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnLeftDoubleClick"), leftModDoubleClickList);

	//Middle
	Fill_ModClickList(GetxRCLineList("*xDeskOnMiddleClickUp"), middleModClickUpList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnMiddleClickDown"), middleModClickDownList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnMiddleDoubleClick"), middleModDoubleClickList);

	//Right
	Fill_ModClickList(GetxRCLineList("*xDeskOnRightClickUp") , rightModClickUpList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnRightClickDown"), rightModClickDownList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnRightDoubleClick"), rightModDoubleClickList);

	// X1
	Fill_ModClickList(GetxRCLineList("*xDeskOnX1ClickUp"), X1ModClickUpList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnX1ClickDown"), X1ModClickDownList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnX1DoubleClick"), X1ModDoubleClickList);

	// X2
	Fill_ModClickList(GetxRCLineList("*xDeskOnX2ClickUp"), X2ModClickUpList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnX2ClickDown"), X2ModClickDownList);
	Fill_ModClickList(GetxRCLineList("*xDeskOnX2DoubleClick"), X2ModDoubleClickList);

	
	//Event Regions
	Fill_ClickRegionList( GetxRCLineList("*xDeskOnLeftClickRegion"), LeftClickRegions);
	Fill_ClickRegionList( GetxRCLineList("*xDeskOnLeftDoubleClickRegion"), LeftDoubleClickRegions);
	
	Fill_ClickRegionList( GetxRCLineList("*xDeskOnMiddleClickRegion"), MiddleClickRegions);
	Fill_ClickRegionList( GetxRCLineList("*xDeskOnMiddleDoubleClickRegion"), MiddleDoubleClickRegions);

	Fill_ClickRegionList( GetxRCLineList("*xDeskOnRightClickRegion"), RightClickRegions);
	Fill_ClickRegionList( GetxRCLineList("*xDeskOnRightDoubleClickRegion"), RightDoubleClickRegions);

	Fill_ClickRegionList( GetxRCLineList("*xDeskOnX1ClickRegion"), X1ClickRegions);
	Fill_ClickRegionList( GetxRCLineList("*xDeskOnX1DoubleClickRegion"), X1DoubleClickRegions);

	Fill_ClickRegionList( GetxRCLineList("*xDeskOnX2ClickRegion"), X2ClickRegions);
	Fill_ClickRegionList( GetxRCLineList("*xDeskOnX2DoubleClickRegion"), X2DoubleClickRegions);

	StringList WorkAreaList = GetxRCLineList("*xDeskWorkArea");

	char monitor[16] = {0};
	char x[16] = {0};
	char y[16] = {0};
	char width[16] = {0};
	char height[16] = {0};
	char *buffers[] = {monitor, x, y, width, height};

	int numTokens = 0;

	SetRectEmpty(&rWorkArea1);
	SetRectEmpty(&rWorkArea2);
	SetRectEmpty(&rWorkArea3);

	for(StringListIterator sit = WorkAreaList.begin(); sit != WorkAreaList.end(); sit++)
	{
		numTokens = LCTokenize((*sit).c_str(), buffers, 5, NULL);
		if(numTokens >= 4)
		{
			if (numTokens == 4)
			{
				rWorkArea1.left = xParseCoord(monitor, 0, iScreenWidth, false);
				rWorkArea1.top = xParseCoord(x, 0, iScreenHeight, true);
				rWorkArea1.right = rWorkArea1.left + xParseDimen(y, 0, iScreenWidth, false);
				rWorkArea1.bottom = rWorkArea1.top + xParseDimen(width, 0, iScreenHeight, true);
				break;
			}
			else if (numTokens > 4)
			{
				if ( _stricmp(monitor, "2") == NULL || _stricmp(monitor, "secondary") == NULL )
				{
					string strX = "2@";
					if (strstr(x, "@") != NULL)
						strX = "";
					strX.append(x);
					string strY = "2@";
					if (strstr(y, "@") != NULL)
						strY = "";
					strY.append(y);
					string strWidth = "2@";
					if (strstr(width, "@") != NULL)
						strWidth = "";
					strWidth.append(width);
					string strHeight = "2@";
					if (strstr(height, "@") != NULL)
						strHeight = "";
					strHeight.append(height);
					rWorkArea2.left = xParseCoord(strX.c_str(), 0, iScreenWidth, false);
					rWorkArea2.top = xParseCoord(strY.c_str(), 0, iScreenHeight, true);
					rWorkArea2.right = rWorkArea2.left + xParseDimen(strWidth.c_str(), 0, iScreenWidth, false);
					rWorkArea2.bottom = rWorkArea2.top + xParseDimen(strHeight.c_str(), 0, iScreenHeight, true);
				}
				else if ( _stricmp(monitor, "3") == NULL || _stricmp(monitor, "third") == NULL )
				{
					string strX = "3@";
					if (strstr(x, "@") != NULL)
						strX = "";
					strX.append(x);
					string strY = "3@";
					if (strstr(y, "@") != NULL)
						strY = "";
					strY.append(y);
					string strWidth = "3@";
					if (strstr(width, "@") != NULL)
						strWidth = "";
					strWidth.append(width);
					string strHeight = "3@";
					if (strstr(height, "@") != NULL)
						strHeight = "";
					strHeight.append(height);
					rWorkArea3.left = xParseCoord(strX.c_str(), 0, iScreenWidth, false);
					rWorkArea3.top = xParseCoord(strY.c_str(), 0, iScreenHeight, true);
					rWorkArea3.right = rWorkArea3.left + xParseDimen(strWidth.c_str(), 0, iScreenWidth, false);
					rWorkArea3.bottom = rWorkArea3.top + xParseDimen(strHeight.c_str(), 0, iScreenHeight, true);
				}
				else
				{
					string strX = "1@";
					if (strstr(x, "@") != NULL)
						strX = "";
					strX.append(x);
					string strY = "1@";
					if (strstr(y, "@") != NULL)
						strY = "";
					strY.append(y);
					string strWidth = "1@";
					if (strstr(width, "@") != NULL)
						strWidth = "";
					strWidth.append(width);
					string strHeight = "1@";
					if (strstr(height, "@") != NULL)
						strHeight = "";
					strHeight.append(height);
					rWorkArea1.left = xParseCoord(strX.c_str(), 0, iScreenWidth, false);
					rWorkArea1.top = xParseCoord(strY.c_str(), 0, iScreenHeight, true);
					rWorkArea1.right = rWorkArea1.left + xParseDimen(strWidth.c_str(), 0, iScreenWidth, false);
					rWorkArea1.bottom = rWorkArea1.top + xParseDimen(strHeight.c_str(), 0, iScreenHeight, true);
				}
			}
		}
	}

	WorkAreaList.clear();

	StringList OverlayList = GetxRCLineList("*xDeskOverlay");

	if ( !OverlayList.empty() )
		bOverlaysUsed = TRUE;

	OverlayList.clear();
					
	SetWorkArea();

	SetWindowLong(hxDeskWnd, GWL_USERDATA, SHELL_MAGICNUMBER);
	SetWindowPos(hxDeskWnd, HWND_BOTTOM, iScreenX, iScreenY, iScreenWidth, iScreenHeight, SWP_NOACTIVATE|SWP_NOSENDCHANGING);
	
	if ( bOverlaysUsed || m_iHueIntensity != 0 || m_iLuminanceIntensity != 0 || m_iSaturationIntensity != 100 || m_iMixIntensity != 0 )
	{
		bOverlaysUsed = TRUE;
		bOverlaysInitialization = TRUE;
		RedrawWindow(hxDeskWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW );
	}

	ShowWindow(hxDeskWnd,SW_SHOWNOACTIVATE);

	StringList DropList = GetxRCLineList("*xDeskOnDrop");

	Fill_ModClickList(DropList, onDropList);

	if ( !DropList.empty() )
		DragAcceptFiles(hxDeskWnd, true);

	DropList.clear();

	return 0;
}

void quitModule(HINSTANCE hInstance)
{
	RemoveBangCommand("!xDeskSetWorkArea");
	RemoveBangCommand("!xDeskSetWallpaper");
	RemoveBangCommand("!xDesk");

	CleanupPointers();

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);
	DestroyWindow(hxDeskWnd);

	if ( backupBitmap )
	{
		backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
		DeleteObject(backupBitmap);
	}
	if ( backupDC )
		DeleteDC(backupDC);
	if ( oldbackupBMP )
		DeleteObject(oldbackupBMP);

	backupDC = NULL;
	backupBitmap = NULL;
	oldbackupBMP = NULL;

	UnregisterClass("xDeskMessageHandler", hInstance);
	UnregisterClass("DesktopBackgroundClass", hInstance);
}

void onWheelDown()
{
	Check_ModClickList(wheelDownList);
}

void onWheelUp()
{
	Check_ModClickList(wheelUpList);
}

void onLButtonDblClk(int x, int y)
{
	if ( !Check_ClickRegionList(x, y, LeftDoubleClickRegions) ) 
		Check_ModClickList(leftModDoubleClickList);
}

void onLButtonDown(int x, int y)
{
	// 1 = LeftButton
	ontemplateButtonDown(x, y, 1, LeftClickRegions);
}

void onLButtonUp(int x, int y)
{	
	// 1 = LeftButton
	ontemplateButtonUp(x, y, 1, LeftClickRegions);
}

void onMButtonDblClk(int x, int y)
{
	if ( !Check_ClickRegionList(x, y, MiddleDoubleClickRegions) ) 
		Check_ModClickList(middleModDoubleClickList);
}

void onMButtonDown(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonDown(x, y, 3, MiddleClickRegions);
}

void onMButtonUp(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonUp(x, y, 3, MiddleClickRegions);
}

void onRButtonDblClk(int x, int y)
{
	if ( !Check_ClickRegionList(x, y, RightDoubleClickRegions) ) 
		Check_ModClickList(rightModDoubleClickList);
}

void onRButtonDown(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonDown(x, y, 2, RightClickRegions);
}

void onRButtonUp(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonUp(x, y, 2, RightClickRegions);
}

void onX1ButtonDblClk(int x, int y)
{
	if ( !Check_ClickRegionList(x, y, X1DoubleClickRegions) ) 
		Check_ModClickList(X1ModDoubleClickList);
}

void onX1ButtonDown(int x, int y)
{
	// 4 = X1Button
	ontemplateButtonDown(x, y, 4, X1ClickRegions);
}

void onX1ButtonUp(int x, int y)
{
	// 4 = X1Button
	ontemplateButtonUp(x, y, 4, X1ClickRegions);
}

void onX2ButtonDblClk(int x, int y)
{
	if ( !Check_ClickRegionList(x, y, X2DoubleClickRegions) ) 
		Check_ModClickList(X2ModDoubleClickList);
}

void onX2ButtonDown(int x, int y)
{
	// 5 = X2Button
	ontemplateButtonDown(x, y, 5, X2ClickRegions);
}

void onX2ButtonUp(int x, int y)
{
	// 5 = X2Button
	ontemplateButtonUp(x, y, 5, X2ClickRegions);
}

void ontemplateButtonDown(int x, int y, int actualMouseClick, RegionList &ClickRegions)
{
	if ( actualMouseClick == 1 )
		Check_ModClickList(leftModClickDownList, -10);
	else if ( actualMouseClick == 2 )
		Check_ModClickList(rightModClickDownList, -20);
	else if ( actualMouseClick == 3 )
		Check_ModClickList(middleModClickDownList, -30);
	else if ( actualMouseClick == 4 )
		Check_ModClickList(X1ModClickDownList);
	else if ( actualMouseClick == 5 )
		Check_ModClickList(X2ModClickDownList);
}

void ontemplateButtonUp(int x, int y, int actualMouseClick, RegionList &ClickRegions)
{
	if ( !Check_ClickRegionList(x, y, ClickRegions) )
	{
		if ( actualMouseClick == 1 )
			Check_ModClickList(leftModClickUpList, -10);
		else if ( actualMouseClick == 2 )
			Check_ModClickList(rightModClickUpList, -20);
		else if ( actualMouseClick == 3 )
			Check_ModClickList(middleModClickUpList, -30);
		else if ( actualMouseClick == 4 )
			Check_ModClickList(X1ModClickUpList);
		else if ( actualMouseClick == 5 )
			Check_ModClickList(X2ModClickUpList);
	}
}

void Fill_ModClickList(StringList inputList, ModClickList &outputList)
{
	if ( inputList.empty() )
		return;

	char modifier[32] = {0};
	char action[MAX_LINE_LENGTH] = {0};
	char *buffers[] = {modifier};

	for(StringListIterator sit = inputList.begin(); sit != inputList.end(); sit++)
	{
		if(LCTokenize((*sit).c_str(), buffers, 1, action) >= 1)
		{
			ModClick *insert = new ModClick();

			if ( _stricmp( modifier, ".none") == NULL || _stricmp( modifier, "none") == NULL )
				insert->iModifier = -1;
			else if ( _stricmp( modifier, ".mouse1") == NULL )
				insert->iModifier = -10;
			else if ( _stricmp( modifier, ".mouse2") == NULL )
				insert->iModifier = -20;
			else if ( _stricmp( modifier, ".mouse3") == NULL )
				insert->iModifier = -30;
			else if ( _stricmp( modifier, ".ctrl") == NULL )
				insert->iModifier = 1;
			else if ( _stricmp( modifier, ".shift") == NULL )
				insert->iModifier = 2;
			else if ( _stricmp( modifier, ".alt") == NULL )
				insert->iModifier = 3;
			else if ( _stricmp( modifier, ".ctrl+shift") == NULL || _stricmp( modifier, ".shift+ctrl") == NULL )
				insert->iModifier = 12;
			else if ( _stricmp( modifier, ".ctrl+alt") == NULL || _stricmp( modifier, ".alt+ctrl") == NULL )
				insert->iModifier = 13;
			else if ( _stricmp( modifier, ".shift+alt") == NULL || _stricmp( modifier, ".alt+shift") == NULL )
				insert->iModifier = 23;
			else if ( _stricmp( modifier, ".ctrl+shift+alt") == NULL 
						|| _stricmp( modifier, ".ctrl+alt+shift") == NULL
						|| _stricmp( modifier, ".shift+alt+ctrl") == NULL
						|| _stricmp( modifier, ".shift+ctrl+alt") == NULL 
						|| _stricmp( modifier, ".alt+ctrl+shift") == NULL
						|| _stricmp( modifier, ".alt+shift+ctrl") == NULL )
				insert->iModifier = 123;
			else
				insert->iModifier = -1;

			insert->strAction = action;
				
			outputList.insert(outputList.begin(), insert);
		}
	}
}

bool Check_ModClickList(ModClickList inputList, int iMouseButton, LONG wParam)
{
	if ( inputList.empty() )
		return false;

	bool found = false;

	for(ModClickListIterator mlit = inputList.begin(); mlit != inputList.end(); mlit++)
	{
		if ( (*mlit)->iModifier != iMouseButton && iMouseButton != 0 && wParam == NULL )
		{
			if ( (*mlit)->iModifier == -10 && bMouse1Pressed )
			{
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
				found = true;
				break;
			}
			else if ( (*mlit)->iModifier == -20 && bMouse2Pressed )
			{
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
				found = true;
				break;
			}
			else if ( (*mlit)->iModifier == -30 && bMouse3Pressed )
			{
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
				found = true;
				break;
			}
		}

		if ( wParam != NULL )
		{
			if ( (*mlit)->iModifier == -20 && GetAsyncKeyState(VK_LBUTTON) < 1 )
			{
				DropExecute((*mlit)->strAction, wParam);
				found = true;
				break;
			}
			else if ( (*mlit)->iModifier == -10 && GetAsyncKeyState(VK_RBUTTON) < 1 )
			{
				DropExecute((*mlit)->strAction, wParam);
				found = true;
				break;
			}
		}

		if ( (*mlit)->iModifier == -1 && !(0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) && !(0x8000 & GetKeyState(VK_MENU)) )
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
		else if ( (*mlit)->iModifier == 123 && (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
		else if ( (*mlit)->iModifier == 12 && (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && !(0x8000 & GetKeyState(VK_MENU)))
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
		else if ( (*mlit)->iModifier == 13 && (0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
		else if ( (*mlit)->iModifier == 23 && !(0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
		else if ( (*mlit)->iModifier == 1 && (0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) && !(0x8000 & GetKeyState(VK_MENU)) )
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
		else if ( (*mlit)->iModifier == 2 && !(0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && !(0x8000 & GetKeyState(VK_MENU)) )
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
		else if ( (*mlit)->iModifier == 3 && !(0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			if ( wParam == NULL )
				LSExecute(NULL, (*mlit)->strAction.c_str(), SW_SHOWNORMAL);
			else
				DropExecute((*mlit)->strAction, wParam);
			found = true;
			break;
		}
	}

	return found;
}

void Fill_ClickRegionList(StringList inputList, RegionList &outputList)
{
	if ( inputList.empty() )
		return;

	char x[16] = {0};
	char y[16] = {0};
	char width[16] = {0};
	char height[16] = {0};
	char command[MAX_LINE_LENGTH] = {0};;
	char *buffers[] = {x, y, width, height};

	for(StringListIterator sit = inputList.begin(); sit != inputList.end(); sit++)
	{
		if(LCTokenize((*sit).c_str(), buffers, 4, command) >= 4)
		{
			Region *insert = new Region();

			insert->strOrgX = x;
			insert->strOrgY = y;
			insert->strOrgWidth = width;
			insert->strOrgHeight = height;

			insert->region.left = xParseCoord(insert->strOrgX.c_str(), 0, iScreenWidth, false)-iScreenX;
			insert->region.right = insert->region.left+xParseDimen(insert->strOrgWidth.c_str(), iScreenWidth, iScreenWidth, false);
			insert->region.top = xParseCoord(insert->strOrgY.c_str(), 0, iScreenHeight, true)-iScreenY;
			insert->region.bottom = insert->region.top+xParseDimen(insert->strOrgHeight.c_str(), iScreenHeight, iScreenHeight, true);

			insert->strAction = command;

			outputList.insert(outputList.begin(), insert);
		}
	}
}

bool Check_ClickRegionList(int x, int y, RegionList inputlist)
{
	bool found = false;

	if (!inputlist.empty())
	{	
		POINT pt = {x,y};
		for(RegionListIterator lrit = inputlist.begin(); lrit != inputlist.end(); lrit++)
		{
			if( PtInRect( &(*lrit)->region, pt) )
			{	
				if ( !(*lrit)->strAction.empty() )
					LSExecute(NULL, (*lrit)->strAction.c_str(), SW_SHOWNORMAL);
			
				found = true;
			}
		}
	}

	return found;
}

void DropExecute(string strAction, LONG wParam)
{
	int intFileAction = -1;

	char action[32] = {0};
	char szTarget[MAX_LINE_LENGTH] = {0};
	char *buffers[] = {action};

	LCTokenize(strAction.c_str(), buffers, 1, szTarget);

	if (_stricmp(action, "move") == NULL)
		intFileAction = 1;
	else if (_stricmp(action, "copy") == NULL)
		intFileAction = 2;
	else if (_stricmp(action, "delete") == NULL)
		intFileAction = 0;
	else if (_stricmp(action, "shortcut") == NULL)
		intFileAction = 3;
	else
		intFileAction = -1;

	if (intFileAction == -1)
	{
		int numDropped, i;
		char szFile[MAX_LINE_LENGTH] = {0};
					
		numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
		for (i = 0; i < numDropped; i++)
		{
			DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_LINE_LENGTH);
			if (szFile && szFile[0])
			{
				CString fullCommand = strAction.c_str();
				if ( fullCommand.Find("%[droppath]%") != -1 )
				{
					fullCommand.Replace("%[droppath]%", szFile);
				}
				else
				{
					fullCommand += " \"";
					fullCommand += szFile;
					fullCommand += "\"";
				}
				LSExecute(NULL, fullCommand, SW_SHOWNORMAL);
			}
		}

		DragFinish((HDROP) wParam);
		return;
	}
			
	//Test Path for Existence
	CString strPath = szTarget;
	strPath.Replace("\"", "");
	if ( strPath.Right(1) != "\\" )
		strPath += "\\";

	BOOL pathExists = PathFileExists(strPath);

	if (intFileAction >= 0 && intFileAction < 3)
	{
		char *szSourceFiles = NULL;
		char buffer[MAX_LINE_LENGTH] = {0};
		int numFiles;
		UINT len,scursor = 0;

		numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

		szSourceFiles = new char[MAX_LINE_LENGTH*numFiles+1];
		scursor = 0;

		for (int i=0;i<numFiles;i++)
		{
			len = DragQueryFile((HDROP) wParam, i, buffer, MAX_LINE_LENGTH);
						
			strcpy(szSourceFiles+scursor,buffer);
			scursor += len + 1;
		}
					
		DragFinish((HDROP) wParam);

		if( pathExists || intFileAction == 0 )
		{
			szSourceFiles[scursor] = 0;
			strcpy(buffer, strPath);
			buffer[strlen(buffer)+1] = 0;
						
			SHFILEOPSTRUCT shfop;
			shfop.hwnd = NULL;

			if (intFileAction == 1)
				shfop.wFunc = FO_MOVE;
			else if (intFileAction == 2)
				shfop.wFunc = FO_COPY;
			else if (intFileAction == 0)
				shfop.wFunc = FO_DELETE;
			else 
				shfop.wFunc = FO_COPY;

			shfop.pFrom = szSourceFiles;
			shfop.pTo = buffer;
			shfop.fFlags = FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO;
			SHFileOperation(&shfop);
		}
					
		delete[] szSourceFiles;
	}
	else if (pathExists && intFileAction == 3)
	{
		int numDropped;
		char szFile[MAX_LINE_LENGTH] = {0};
					
		numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
		for (int i = 0; i < numDropped; i++)
		{
			DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_LINE_LENGTH);

			HRESULT hres; 
			IShellLink* psl; 

			hres = CoCreateInstance(CLSID_ShellLink, NULL,
			CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID *) &psl); 
			if (SUCCEEDED(hres))
			{
				IPersistFile* ppf; 
				psl->SetPath(szFile); 
				psl->SetDescription("");
 
				hres = psl->QueryInterface(IID_IPersistFile, (LPVOID*)&ppf); 
				if (SUCCEEDED(hres))
				{
					CString work = szFile;
					work = work.Right(work.GetLength()-work.ReverseFind('\\')-1);
					work = work.Left(work.Find("."));
					work += ".lnk";

					WCHAR wsz[MAX_PATH]; 
					MultiByteToWideChar(CP_ACP, 0, (strPath + work), -1, wsz, MAX_LINE_LENGTH);  
					hres = ppf->Save(wsz, TRUE); 
					ppf->Release(); 
				}
				psl->Release(); 
			}
		}
		DragFinish((HDROP) wParam);
	}
}

void CleanupPointers()
{
	RegionListIterator llit;

	//Left Click Regions
	for(llit = LeftClickRegions.begin(); llit != LeftClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	LeftClickRegions.clear();
	
	for(llit = LeftDoubleClickRegions.begin(); llit != LeftDoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	LeftDoubleClickRegions.clear();
	
	//Middle Click Regions
	for(llit = MiddleClickRegions.begin(); llit != MiddleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	MiddleClickRegions.clear();

	for(llit = MiddleDoubleClickRegions.begin(); llit != MiddleDoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	MiddleDoubleClickRegions.clear();

	//Right Click Regions
	for(llit = RightClickRegions.begin(); llit != RightClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	RightClickRegions.clear();

	for(llit = RightDoubleClickRegions.begin(); llit != RightDoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	RightDoubleClickRegions.clear();

	//X1 Click Regions
	for(llit = X1ClickRegions.begin(); llit != X1ClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	X1ClickRegions.clear();

	for(llit = X1DoubleClickRegions.begin(); llit != X1DoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	X1DoubleClickRegions.clear();

	//X2 Click Regions
	for(llit = X2ClickRegions.begin(); llit != X2ClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	X2ClickRegions.clear();

	for(llit = X2DoubleClickRegions.begin(); llit != X2DoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	X2DoubleClickRegions.clear();

	//Delete All ModClick Pointers
	//------------------------------------------------------------------------------------------------
	ModClickListIterator mlit;

	// Drop
	for(mlit = onDropList.begin(); mlit != onDropList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	onDropList.clear();

	// WheelUp
	for(mlit = wheelUpList.begin(); mlit != wheelUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	wheelUpList.clear();

	// WheelDown
	for(mlit = wheelDownList.begin(); mlit != wheelDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	wheelDownList.clear();

	// Left
	for(mlit = leftModClickUpList.begin(); mlit != leftModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	leftModClickUpList.clear();

	for(mlit = leftModClickDownList.begin(); mlit != leftModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	leftModClickDownList.clear();

	for(mlit = leftModDoubleClickList.begin(); mlit != leftModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	leftModDoubleClickList.clear();

	// Middle
	for(mlit = middleModClickUpList.begin(); mlit != middleModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	middleModClickUpList.clear();

	for(mlit = middleModClickDownList.begin(); mlit != middleModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	middleModClickDownList.clear();

	for(mlit = middleModDoubleClickList.begin(); mlit != middleModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	middleModDoubleClickList.clear();

	// Right
	for(mlit = rightModClickUpList.begin(); mlit != rightModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	rightModClickUpList.clear();

	for(mlit = rightModClickDownList.begin(); mlit != rightModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	rightModClickDownList.clear();

	for(mlit = rightModDoubleClickList.begin(); mlit != rightModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	rightModDoubleClickList.clear();

	// X1
	for(mlit = X1ModClickUpList.begin(); mlit != X1ModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X1ModClickUpList.clear();

	for(mlit = X1ModClickDownList.begin(); mlit != X1ModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X1ModClickDownList.clear();

	for(mlit = X1ModDoubleClickList.begin(); mlit != X1ModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X1ModDoubleClickList.clear();

	// X2
	for(mlit = X2ModClickUpList.begin(); mlit != X2ModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X2ModClickUpList.clear();

	for(mlit = X2ModClickDownList.begin(); mlit != X2ModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X2ModClickDownList.clear();

	for(mlit = X2ModDoubleClickList.begin(); mlit != X2ModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X2ModDoubleClickList.clear();
}

StringList GetxRCLineList(const string &uniqueKey)
{
	StringList resultList;
	LPVOID f = LCOpen(NULL);

	if(f != NULL)
	{
		char lineBuffer[MAX_LINE_LENGTH];
		char temp[64], configline[MAX_LINE_LENGTH];
		char *buffers[] = {temp};

		while(LCReadNextConfig(f, uniqueKey.c_str(), lineBuffer, MAX_LINE_LENGTH))
		{
			if (LCTokenize(lineBuffer, buffers, 1, configline) >= 1)
				resultList.insert(resultList.end(), (string) configline);
		}
		
		LCClose(f);
	}

	return resultList;
}

void SetWorkArea()
{
	if ( IsRectEmpty(&rWorkArea1) )
	{
		if ( iRealMonitors > 1 )
		{
			rWorkArea1.left = xParseCoord("1@0", 0, iScreenWidth, false);
			rWorkArea1.top = xParseCoord("1@0", 0, iScreenHeight, true);
			rWorkArea1.right = rWorkArea1.left + xParseDimen("1@100%", 0, iScreenWidth, false);
			rWorkArea1.bottom = rWorkArea1.top + xParseDimen("1@100%", 0, iScreenHeight, true);

			if ( !IsRectEmpty(&rWorkArea2)  )
			{
				rWorkArea2.left = xParseCoord("2@0", 0, iScreenWidth, false);
				rWorkArea2.top = xParseCoord("2@0", 0, iScreenHeight, true);
				rWorkArea2.right = rWorkArea2.left + xParseDimen("2@100%", 0, iScreenWidth, false);
				rWorkArea2.bottom = rWorkArea2.top + xParseDimen("2@100%", 0, iScreenHeight, true);
			}

			if ( iRealMonitors > 2 && !IsRectEmpty(&rWorkArea3)  )
			{
				rWorkArea3.left = xParseCoord("3@0", 0, iScreenWidth, false);
				rWorkArea3.top = xParseCoord("3@0", 0, iScreenHeight, true);
				rWorkArea3.right = rWorkArea3.left + xParseDimen("3@100%", 0, iScreenWidth, false);
				rWorkArea3.bottom = rWorkArea3.top + xParseDimen("3@100%", 0, iScreenHeight, true);
			}
	
			SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea1, SPIF_SENDCHANGE);
			SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea2, SPIF_SENDCHANGE);
			if ( iRealMonitors > 2 )
				SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea3, SPIF_SENDCHANGE);
		}
		else
		{
			rWorkArea1.left = 0;
			rWorkArea1.right = iScreenWidth;
			rWorkArea1.top = 0;
			rWorkArea1.bottom = iScreenHeight;

			SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea1, SPIF_SENDCHANGE);
		}
	}
	else
	{
		SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea1, SPIF_SENDCHANGE);

		if ( !IsRectEmpty(&rWorkArea2) && iRealMonitors > 1 )
			SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea2, SPIF_SENDCHANGE);
		if ( !IsRectEmpty(&rWorkArea3) && iRealMonitors > 2 )
			SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea3, SPIF_SENDCHANGE);
	}

	bIsFirst = FALSE;
}

void BangSetWorkArea(HWND hwndCaller, LPCSTR pszArgs)
{
	char monitor[16] = {0};
	char x[16] = {0};
	char y[16] = {0};
	char width[16] = {0};
	char height[16] = {0};
	char *buffers[] = {monitor, x, y, width, height};

	int numTokens = LCTokenize(pszArgs, buffers, 5, NULL);
	
	if(numTokens >= 4)
	{
		if (numTokens == 4)
		{
			rWorkArea1.left = xParseCoord(monitor, 0, iScreenWidth, false);
			rWorkArea1.top = xParseCoord(x, 0, iScreenHeight, true);
			rWorkArea1.right = rWorkArea1.left + xParseDimen(y, 0, iScreenWidth, false);
			rWorkArea1.bottom = rWorkArea1.top + xParseDimen(width, 0, iScreenHeight, true);

			SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea1, SPIF_SENDCHANGE);
		}
		else if (numTokens > 4)
		{
			if ( _stricmp(monitor, "2") == NULL || _stricmp(monitor, "secondary") == NULL )
			{
				string strX = "2@";
				if (strstr(x, "@") != NULL)
					strX = "";
				strX.append(x);
				string strY = "2@";
				if (strstr(y, "@") != NULL)
					strY = "";
				strY.append(y);
				string strWidth = "2@";
				if (strstr(width, "@") != NULL)
					strWidth = "";
				strWidth.append(width);
				string strHeight = "2@";
				if (strstr(height, "@") != NULL)
					strHeight = "";
				strHeight.append(height);
				rWorkArea2.left = xParseCoord(strX.c_str(), 0, iScreenWidth, false);
				rWorkArea2.top = xParseCoord(strY.c_str(), 0, iScreenHeight, true);
				rWorkArea2.right = rWorkArea2.left + xParseDimen(strWidth.c_str(), 0, iScreenWidth, false);
				rWorkArea2.bottom = rWorkArea2.top + xParseDimen(strHeight.c_str(), 0, iScreenHeight, true);

				SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea2, SPIF_SENDCHANGE);
			}
			else if ( _stricmp(monitor, "3") == NULL || _stricmp(monitor, "third") == NULL )
			{
				string strX = "3@";
				if (strstr(x, "@") != NULL)
					strX = "";
				strX.append(x);
				string strY = "3@";
				if (strstr(y, "@") != NULL)
					strY = "";
				strY.append(y);
				string strWidth = "3@";
				if (strstr(width, "@") != NULL)
					strWidth = "";
				strWidth.append(width);
				string strHeight = "3@";
				if (strstr(height, "@") != NULL)
					strHeight = "";
				strHeight.append(height);
				rWorkArea3.left = xParseCoord(strX.c_str(), 0, iScreenWidth, false);
				rWorkArea3.top = xParseCoord(strY.c_str(), 0, iScreenHeight, true);
				rWorkArea3.right = rWorkArea3.left + xParseDimen(strWidth.c_str(), 0, iScreenWidth, false);
				rWorkArea3.bottom = rWorkArea3.top + xParseDimen(strHeight.c_str(), 0, iScreenHeight, true);

				SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea3, SPIF_SENDCHANGE);
			}
			else
			{
				string strX = "1@";
				if (strstr(x, "@") != NULL)
					strX = "";
				strX.append(x);
				string strY = "1@";
				if (strstr(y, "@") != NULL)
					strY = "";
				strY.append(y);
				string strWidth = "1@";
				if (strstr(width, "@") != NULL)
					strWidth = "";
				strWidth.append(width);
				string strHeight = "1@";
				if (strstr(height, "@") != NULL)
					strHeight = "";
				strHeight.append(height);
				rWorkArea1.left = xParseCoord(strX.c_str(), 0, iScreenWidth, false);
				rWorkArea1.top = xParseCoord(strY.c_str(), 0, iScreenHeight, true);
				rWorkArea1.right = rWorkArea1.left + xParseDimen(strWidth.c_str(), 0, iScreenWidth, false);
				rWorkArea1.bottom = rWorkArea1.top + xParseDimen(strHeight.c_str(), 0, iScreenHeight, true);

				SystemParametersInfo(SPI_SETWORKAREA, (bIsFirst ? 0:1), &rWorkArea1, SPIF_SENDCHANGE);
			}
		}
	}
}

void BangSetWallpaper(HWND hwndCaller, LPCSTR pszArgs)
{
	string strArgs = pszArgs;	
	
	char szThemePath[MAX_LINE_LENGTH] = {0};

	if ( GetxRCBool("themedir", "") )
		VarExpansion( szThemePath, "$themedir$" );
	else
		VarExpansion( szThemePath, "$litestepdir$" );

	strcat(szThemePath, "wallpaper.bmp");

	if ( strArgs.empty() )
	{
		OverlayPainter(NULL);

		if ( SaveBitmap(backupDC, iScreenWidth, iScreenHeight, szThemePath) == 1 )
		{
			// Change the wallpaper.
			SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, szThemePath, SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE );  
		}
	}
	else 
	{
		PaintTexture *pTexture = Create_xTextureClass();

		char file[128] = {0};
		char *buffers[] = {file};

		LCTokenize(strArgs.c_str(), buffers, 1, NULL);

		strArgs = file;
		
		LSSetVariable("xDeskWallpaperPaintingMode", ".image");
		LSSetVariable("xDeskWallpaperImage", ("\"" + strArgs + "\"").c_str());
		LSSetVariable("xDeskWallpaperImageMode", ".stretch_quadratic");

		pTexture->configure("xDesk", "Wallpaper");
	
		HDC tmpMemoryDC = CreateCompatibleDC(NULL);

		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=iScreenWidth;      // Set size you need
		alpha.bmiHeader.biHeight=iScreenHeight;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		HBITMAP tmpMemoryBitmap = CreateDIBSection( tmpMemoryDC,
											&alpha,
											DIB_RGB_COLORS,
											&pvBits,
											NULL,
											0 );
		
		HBITMAP oldtmpMemoryBMP = (HBITMAP) SelectObject(tmpMemoryDC, tmpMemoryBitmap);

		pTexture->apply(tmpMemoryDC, 0, 0, iScreenWidth, iScreenHeight);

		if ( SaveBitmap(tmpMemoryDC, iScreenWidth, iScreenHeight, szThemePath) == 1 )
		{
			// Change the wallpaper.
			SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, szThemePath, SPIF_UPDATEINIFILE | SPIF_SENDWININICHANGE ); 
		}

		Destroy_xTextureClass(pTexture);
		pTexture = NULL;

		tmpMemoryBitmap = (HBITMAP) SelectObject(tmpMemoryDC, oldtmpMemoryBMP);
		DeleteDC(tmpMemoryDC);	
		DeleteObject(tmpMemoryBitmap);
		DeleteObject(oldtmpMemoryBMP);
	}
}

void BangSetup(HWND hwndCaller, LPCSTR pszArgs)
{
	char setting[64] = {0};
	char action[MAX_LINE_LENGTH] = {0};
	char *buffers[] = {setting};

	LCTokenize(pszArgs, buffers, 1, action);

	StringList modifyList;
	modifyList.insert(modifyList.end(), action);

	//Wheel
	if ( _stricmp("OnWheelDown", setting) == NULL )
		Fill_ModClickList(modifyList, wheelDownList);
	else if ( _stricmp("OnWheelUp", setting) == NULL )
		Fill_ModClickList(modifyList, wheelUpList);

	//Left
	else if ( _stricmp("OnLeftClickUp", setting) == NULL )
		Fill_ModClickList(modifyList, leftModClickUpList);
	else if ( _stricmp("OnLeftClickDown", setting) == NULL )
		Fill_ModClickList(modifyList, leftModClickDownList);
	else if ( _stricmp("OnLeftDoubleClick", setting) == NULL )
		Fill_ModClickList(modifyList, leftModDoubleClickList);

	//Middle
	else if ( _stricmp("OnMiddleClickUp", setting) == NULL )
		Fill_ModClickList(modifyList, middleModClickUpList);
	else if ( _stricmp("OnMiddleClickDown", setting) == NULL )
		Fill_ModClickList(modifyList, middleModClickDownList);
	else if ( _stricmp("OnMiddleDoubleClick", setting) == NULL )
		Fill_ModClickList(modifyList, middleModDoubleClickList);

	//Right
	else if ( _stricmp("OnRightClickUp", setting) == NULL )
		Fill_ModClickList(modifyList, rightModClickUpList);
	else if ( _stricmp("OnRightClickDown", setting) == NULL )
		Fill_ModClickList(modifyList, rightModClickDownList);
	else if ( _stricmp("OnRightDoubleClick", setting) == NULL )
		Fill_ModClickList(modifyList, rightModDoubleClickList);

	// X1
	else if ( _stricmp("OnX1ClickUp", setting) == NULL )
		Fill_ModClickList(modifyList, X1ModClickUpList);
	else if ( _stricmp("OnX1ClickDown", setting) == NULL )
		Fill_ModClickList(modifyList, X1ModClickDownList);
	else if ( _stricmp("OnX1DoubleClick", setting) == NULL )
		Fill_ModClickList(modifyList, X1ModDoubleClickList);

	// X2
	else if ( _stricmp("OnX2ClickUp", setting) == NULL )
		Fill_ModClickList(modifyList, X2ModClickUpList);
	else if ( _stricmp("OnX2ClickDown", setting) == NULL )
		Fill_ModClickList(modifyList, X2ModClickDownList);
	else if ( _stricmp("OnX2DoubleClick", setting) == NULL )
		Fill_ModClickList(modifyList, X2ModDoubleClickList);

	//Event Regions
	else if ( _stricmp("OnLeftClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, LeftClickRegions);
	else if ( _stricmp("OnLeftDoubleClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, LeftDoubleClickRegions);
	
	else if ( _stricmp("OnMiddleClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, MiddleClickRegions);
	else if ( _stricmp("OnMiddleDoubleClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, MiddleDoubleClickRegions);

	else if ( _stricmp("OnRightClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, RightClickRegions);
	else if ( _stricmp("OnRightDoubleClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, RightDoubleClickRegions);

	else if ( _stricmp("OnX1ClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, X1ClickRegions);
	else if ( _stricmp("OnX1DoubleClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, X1DoubleClickRegions);

	else if ( _stricmp("OnX2ClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, X2ClickRegions);
	else if ( _stricmp("OnX2DoubleClickRegion", setting) == NULL )
		Fill_ClickRegionList( modifyList, X2DoubleClickRegions);
}

void OverlayPainter(HDC hDC)
{
	if ( backupBitmap == NULL && backupDC == NULL )
	{
		backupDC = CreateCompatibleDC(NULL);
								
		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=iScreenWidth;      // Set size you need
		alpha.bmiHeader.biHeight=iScreenHeight;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		backupBitmap = CreateDIBSection( backupDC,
											&alpha,
											DIB_RGB_COLORS,
											&pvBits,
											NULL,
											0 );

		oldbackupBMP = (HBITMAP) SelectObject(backupDC, backupBitmap);

		PaintDesktopEx(backupDC, 0, 0, iScreenWidth, iScreenHeight, 0, 0, false);

		PrepareImageEffects();

		char cone[256] = {0};
		char ctwo[256] = {0};
		char cthree[256] = {0};
		char cfour[256] = {0};
		char cfive[256] = {0};
		char csix[256] = {0};
		char *cbuffers[] = {cone, ctwo, cthree, cfour, cfive, csix};
							
		int numTokens = 0;
						
		StringList OverlayList = GetxRCLineList("*xDeskOverlay");
						
		PaintTexture *pTexture = NULL; 

		for (StringListIterator sit = OverlayList.begin(); sit != OverlayList.end(); sit++)
		{
			numTokens = LCTokenize( (*sit).c_str(), cbuffers, 6, NULL);

			if ( numTokens >= 5 )
			{
				pTexture =  Create_xTextureClass();

				if ( numTokens >= 6 && IsOS( OS_2KXP ) && (strchr(csix, 'a') != 0 || strchr(csix, 'A') != 0) )
					pTexture->configure(cfive, "", true);
				else 
					pTexture->configure(cfive, "", false);
						
				pTexture->apply(backupDC, xParseCoord(cone, 0, iScreenWidth, false)-iScreenX, xParseCoord(ctwo, 0, iScreenHeight, true)-iScreenY, xParseDimen(cthree, 0, iScreenWidth, false), xParseDimen(cfour, 0, iScreenHeight, true));

				Destroy_xTextureClass(pTexture);
				pTexture = NULL;
			}
		}

		OverlayList.clear();
		
		if ( hDC != NULL )
			BitBlt(hDC, 0, 0, iScreenWidth, iScreenHeight, backupDC, 0, 0, SRCCOPY);
	}
	else if ( hDC != NULL )
	{
		BitBlt(hDC, 0, 0, iScreenWidth, iScreenHeight, backupDC, 0, 0, SRCCOPY);
	}
}

void PrepareImageEffects()
{
	if ( !backupBitmap || (m_iHueIntensity == 0 && m_iLuminanceIntensity == 0 && m_iSaturationIntensity == 100 && m_iMixIntensity == 0) )
		return;

	//Paint Saturation and Hueing Effects in MemoryDC's
	BITMAP bm;
	GetObject(backupBitmap, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	BYTE r = 0;
	BYTE g = 0;
	BYTE b = 0;

	BYTE h = 0;
	BYTE s = 0;
	BYTE l = 0;
	
	BYTE hueh = 0;
	BYTE hues = 0;
	BYTE huel = 0;

	bool needHSLConversion = false;
	bool needMixColor = false;

	if ( m_iLuminanceIntensity != 0 || m_iSaturationIntensity != 100 )
		needHSLConversion = true;

	if ( m_iHueIntensity != 0 )
	{
		r = GetRValue(m_clrHueColor);
		g = GetGValue(m_clrHueColor);
		b = GetBValue(m_clrHueColor);

		if ( r == b && b == g )
		{
			m_iHueIntensity = 0;
		}
		else
		{
			RGB2HSL(b, g, r, hueh, hues, huel); //Switch r and b due to AlphaMap/bm_b.bmHeight * bm_b.bmWidth
			needHSLConversion = true;
		}
	}

	if ( m_iMixIntensity != 0 )
	{
		m_iMixIntensity = (int) min(255, m_iMixIntensity*255/100);
		needMixColor = true;
	}

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int i = 0; i < iSize; ++i)
	{
		pPixel[3] = 255;

		r = pPixel[0];
		g = pPixel[1];
		b = pPixel[2];

		if ( needHSLConversion )
		{
			//AlphaMap Colorizing
			RGB2HSL(r, g, b, h, s, l);

			if(m_iHueIntensity != 0)
				h = (BYTE) max(0, min(240, (h*(100-m_iHueIntensity)+hueh*m_iHueIntensity)/100));
		
			if(m_iLuminanceIntensity != 0)
				l = (BYTE) max(0, min(240, ((l*100)+(120*m_iLuminanceIntensity))/100));

			if(m_iSaturationIntensity != 100)
			{
				if ( m_iSaturationIntensity > 100 )
					s = (BYTE) max(0, min(240, (240*(m_iSaturationIntensity-100))/100));				
				else
					s = (BYTE) max(0, min(240, (s*m_iSaturationIntensity)/100));
			}

			HSL2RGB(r, g, b, h, s, l);
		}

		if ( needMixColor )
		{
			// HueMix effect
			r = (BYTE)((r * (255 - m_iMixIntensity) + GetBValue(m_clrMixColor) * m_iMixIntensity + 255) >> 8);
			g = (BYTE)((g * (255 - m_iMixIntensity) + GetGValue(m_clrMixColor) * m_iMixIntensity + 255) >> 8);
			b = (BYTE)((b * (255 - m_iMixIntensity) + GetRValue(m_clrMixColor) * m_iMixIntensity + 255) >> 8);
		}

		pPixel[0] = r;
		pPixel[1] = g;
		pPixel[2] = b;

		pPixel+= 4;
	}
}

void RGB2HSL(unsigned char R,unsigned char G,unsigned char B,unsigned char &H,unsigned char &S,unsigned char &L)
{
	float R2, G2, B2, H2, S2, L2;

	//First scale down...
	R2 = ((float)R) / 255.0f;
	G2 = ((float)G) / 255.0f;
	B2 = ((float)B) / 255.0f;

	//...then convert with float...
	RGB2HSLFloat(R2, G2, B2, H2, S2, L2);

	//...and then rescale to 0-255.
	H = (unsigned char)(max(0, min((H2 * 255.0) / 359.0, 255)) + 0.5);
	S = (unsigned char)(max(0, min(S2 * 255.0, 255)) + 0.5);
	L = (unsigned char)(max(0, min(L2 * 255.0, 255)) + 0.5);
}

// Do a precise HSL -> RGB conversion by using
// the float functions and rescale the values
void HSL2RGB(unsigned char &R,unsigned char &G,unsigned char &B,unsigned char H,unsigned char S,unsigned char L)
{
	float R2, G2, B2, H2, S2, L2;

	//First scale down...
	H2 = (((float)H) * 359.0f) / 255.0f;
	S2 = ((float)S) / 255.0f;
	L2 = ((float)L) / 255.0f;

	//...then convert with float...
	HSL2RGBFloat(R2, G2, B2, H2, S2, L2);

	//...and then rescale to 0-255.
	R = (unsigned char)(max(0, min(R2 * 255.0, 255)) + 0.5);
	G = (unsigned char)(max(0, min(G2 * 255.0, 255)) + 0.5);
	B = (unsigned char)(max(0, min(B2 * 255.0, 255)) + 0.5);
}

//Helper-Function for color conversions
//Returns the Intensity of either R, G or B
//depending on its input
float ToRGB(float rm1, float rm2, float rh)
{
	if      (rh > 360.0f) rh -= 360.0f;
	else if (rh <   0.0f) rh += 360.0f;

	if      (rh <  60.0f) rm1 = rm1 + (rm2 - rm1) * rh / 60.0f;   
	else if (rh < 180.0f) rm1 = rm2;
	else if (rh < 240.0f) rm1 = rm1 + (rm2 - rm1) * (240.0f - rh) / 60.0f;      
	                
	return rm1;
}

//Convert a Color: RGB -> HSL
//input and output are floats, so range from RGB and HSL is not 0-255!!
//Important: Range is 0-1  but for H it is 0-360 (all with commas!)!!!!
void RGB2HSLFloat(float R, float G, float B, float &H, float &S, float &L)
{
/*#ifdef COLOR_SAFE //in safe mode evaluate the input
    R = min(max(0.0f, R), 1.0f);
	G = min(max(0.0f, G), 1.0f);
	B = min(max(0.0f, B), 1.0f);
#endif*/

	float minval, maxval, mdiff, msum;
	float rnorm, gnorm, bnorm;
    
	//Get the minimum of R, G, B
	minval = min(min(R, G), B);
	//Get the maximum of R, G, B
	maxval = max(max(R, G), B);

	mdiff  = maxval - minval;
    msum   = maxval + minval;
   
	//L(uminance) is the average of min and max
    L = msum / 2.0f;

    if (msum == 0.0f) 
    {
		//if min == max -> a grey color
		S = 0.0f;
		H = 0.0f; 
    }   
    else 
    { 
		//normalize the colors
		rnorm = (maxval - R) / mdiff;      
		gnorm = (maxval - G) / mdiff;
		bnorm = (maxval - B) / mdiff;   

		//Calculate S
		if(L <= 0.5f) S = mdiff / msum;
		else S = mdiff / (2.0f - msum);

		//Calculate H
		if (R == maxval) H = 60.0f * (6.0f + bnorm - gnorm);
		else if (G == maxval) H = 60.0f * (2.0f + rnorm - bnorm);
		else if (B == maxval) H = 60.0f * (4.0f + gnorm - rnorm);

		//do a modulo of 360 on float
		if (H > 360.0f) H = H - 360.0f;
    }
}

//Convert a Color: HSL -> RGB
//input and output are floats, so range from RGB and HSL is not 0-255!!
//Important: Range is 0-1 (with commas!) !!!!
void HSL2RGBFloat(float &R, float &G, float &B, float H, float S, float L)
{
/*#ifdef COLOR_SAFE //in safe mode evaluate the input
    H = min(max(0.0f, H), 360.0f);
	S = min(max(0.0f, S), 1.0f);
	L = min(max(0.0f, L), 1.0f);
#endif*/

	float rm1, rm2;

	//No saturation -> R, G, B is equal to L
	if(S == 0.0) R = G = B = L;
	else
	{
		if(L <= 0.5f) rm2 = L + L * S;  
		else rm2 = L + S - L * S;
		
		rm1 = 2.0f * L - rm2;
		
		//Convert H(ue) to R, G, B
		R = ToRGB(rm1, rm2, H + 120);
		G = ToRGB(rm1, rm2, H);
		B = ToRGB(rm1, rm2, H - 120);
	}
}

int SaveBitmap(HDC hdc, int width, int height, string strFilename)
{
    void *dibvalues;
    
    DWORD bytes_write;
    DWORD bytes_written;

    HDC hdc2 = CreateCompatibleDC(hdc);
     
	BITMAPINFOHEADER bmih;
    ZeroMemory(&bmih,sizeof(BITMAPINFOHEADER));
    bmih.biSize=sizeof(BITMAPINFOHEADER);
    bmih.biHeight=height;
    bmih.biWidth=width;
    bmih.biPlanes=1;
    bmih.biBitCount=24;
    bmih.biCompression=BI_RGB;
    bmih.biSizeImage = ((((bmih.biWidth * bmih.biBitCount) + 31) & ~31) >> 3) * bmih.biHeight;
    bmih.biXPelsPerMeter = 0;
    bmih.biYPelsPerMeter = 0;
    bmih.biClrImportant = 0;
   
	BITMAPINFO bi;
    bi.bmiHeader = bmih;

    HBITMAP aBmp = CreateDIBSection(hdc,&bi,DIB_RGB_COLORS,(void**)&dibvalues,NULL,NULL);
       
    if (aBmp == NULL)
	{
		DeleteDC(hdc2);
        return 0;
	}

    HGDIOBJ OldObj = SelectObject(hdc2,aBmp);
    
	BitBlt(hdc2,0,0,width,height,hdc,0,0,SRCCOPY);

	BITMAPFILEHEADER bmfh;
    ZeroMemory(&bmfh,sizeof(BITMAPFILEHEADER));
    bmfh.bfOffBits=sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
    bmfh.bfSize=(3*bmih.biHeight*bmih.biWidth)+sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
    bmfh.bfType=0x4d42;
    bmfh.bfReserved1 = 0;
    bmfh.bfReserved2 = 0;

    HANDLE fileHandle = CreateFile(strFilename.c_str(),GENERIC_READ | GENERIC_WRITE,(DWORD)0,NULL,
                                                    CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,(HANDLE) NULL);

    if (fileHandle == INVALID_HANDLE_VALUE)
    {
		DeleteObject(SelectObject(hdc2,OldObj));
		DeleteDC(hdc2);
        return 0;
    }

        // Write the BITMAPFILEHEADER
    bytes_write = sizeof(BITMAPFILEHEADER);
    if (!WriteFile(fileHandle,(void*)&bmfh,bytes_write,&bytes_written,NULL))
    {
		DeleteObject(SelectObject(hdc2,OldObj));
		DeleteDC(hdc2);
        return 0;
    }

        //Write the BITMAPINFOHEADER
    bytes_write = sizeof(BITMAPINFOHEADER);
    if (!WriteFile(fileHandle,(void*)&bmih,bytes_write,&bytes_written,NULL))
    {
		DeleteObject(SelectObject(hdc2,OldObj));
		DeleteDC(hdc2);
        return 0;
    }

        //Write the Color Index Array???
    bytes_write = bmih.biSizeImage;//3*bmih.biHeight*bmih.biWidth;
    if (!WriteFile(fileHandle,(void*)dibvalues,bytes_write,&bytes_written,NULL))
    {
		DeleteObject(SelectObject(hdc2,OldObj));
		DeleteDC(hdc2);
        return 0;
    }

    CloseHandle(fileHandle);

    DeleteObject(SelectObject(hdc2,OldObj));
    DeleteDC(hdc2);

    return 1;
}

// retrieve a string value from the registry
LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue)
{
	DWORD dwDataType;
	DWORD dwDataSize = (DWORD) nBufferSize;
	HKEY hSubKey = NULL;
	LONG lErrorCode;
	
	if(pszSubKey)
	{
		lErrorCode = RegOpenKeyEx(hKey, pszSubKey, 0, KEY_QUERY_VALUE, &hSubKey);
		
		if(lErrorCode != ERROR_SUCCESS)
		{
			StringCchCopy(pszBuffer, nBufferSize, pszDefaultValue);
			return pszBuffer;
		}
		
		hKey = hSubKey;
	}
	
	lErrorCode = RegQueryValueEx(hKey, pszValueName, NULL, &dwDataType,
		(LPBYTE) pszBuffer, &dwDataSize);
	
	if(hSubKey)
		RegCloseKey(hSubKey);
		
	if(dwDataType != REG_SZ || lErrorCode != ERROR_SUCCESS)
		StringCchCopy(pszBuffer, nBufferSize, pszDefaultValue);
	
	return pszBuffer;
}